from typing import List, Optional, Tuple

import torch
import torch.distributions as td
from torch.nn import Parameter, ParameterList

from causica.distributions.noise_accessible.noise_accessible import NoiseAccessible
from causica.distributions.splines.rational_quadratic_transform import PiecewiseRationalQuadraticTransform

SplineParamListType = List[Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]]


class SplineDistribution(td.TransformedDistribution, NoiseAccessible):
    """A Spline Based Noise Distribution."""

    def __init__(
        self,
        base_loc: torch.Tensor,
        base_scale: torch.Tensor,
        param_list: SplineParamListType,
        output_bias: Optional[torch.Tensor] = None,
    ):
        self.base_loc = base_loc
        self.base_scale = base_scale
        self.param_list = param_list

        output_affine_transform: List[td.Transform] = []
        if output_bias is not None:
            output_affine_transform = [
                td.AffineTransform(loc=output_bias, scale=torch.ones_like(output_bias, device=output_bias.device))
            ]

        super().__init__(
            base_distribution=td.Normal(loc=self.base_loc, scale=self.base_scale),
            transforms=[_create_composite_layer(*layer) for layer in param_list] + output_affine_transform,
        )

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        output_affine_transform = self.transforms[-1]
        return output_affine_transform.inv(samples)

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        output_affine_transform = self.transforms[-1]
        return output_affine_transform(noise)


def create_spline_dist_params(
    features: int, flow_steps: int = 1, init_scale: float = 1e-2, num_bins: int = 8
) -> ParameterList:
    param_list = ParameterList()
    for i in range(flow_steps + 1):
        scale = Parameter(
            torch.zeros(features), requires_grad=True
        )  # this will be exponentiated when passed to the spline distribution
        loc = Parameter(torch.zeros(features), requires_grad=True)
        if i == flow_steps:
            knot_locations = None
            derivatives = None
        else:
            knot_locations = Parameter(init_scale * torch.randn(features, num_bins, 2))
            derivatives = Parameter(init_scale * torch.randn(features, num_bins - 1))
        param_list.append(ParameterList([loc, scale, knot_locations, derivatives]))
    return param_list


def _create_composite_layer(
    loc: torch.Tensor,
    scale: torch.Tensor,
    knot_locations: Optional[torch.Tensor],
    derivatives: Optional[torch.Tensor],
) -> td.Transform:
    """Create the composite layers as required by the spline distribution."""
    affine = td.AffineTransform(loc=loc, scale=scale.exp())
    if knot_locations is None:
        return affine
    return td.ComposeTransform([affine, PiecewiseRationalQuadraticTransform(knot_locations, derivatives)])



from typing import Callable

import torch.distributions as td
from torch import nn


class ParametrizedDistribution(nn.Module):
    """
    A container class for a distribution and its parameters.

    In PyTorch, `Distribution`s are not learnable as the parameters are inputs to the class.
    This can mean that after updates, if any other values within the distribution are derived from the parameters,
    they may not be updated.

    This simple wrapper stores a distribution class and its parameters (as a dict) together.
    It dynamically instantiates the distribution when required. It is assumed that distribution instantiation is cheap.
    For the reason above, the dynamically instantiated distribution should be persisted for as little time as possible (particularly if parameters are updated).

    For example:
    ```
    loc = Parameter(0., requires_grad=True)
    scale = Parameter(1., requires_grad=True)
    paramdist = ParametrizedDistribution(
        torch.distributions.Normal,
        torch.nn.ParameterDict(dict(loc=loc, scale=scale)),
    )
    ```
    And then to use:
    ```
    dist: torch.distributions.Normal = paramdist.forward()
    mean = dist.mean
    ```
    """

    def __init__(self, distribution_class: Callable[..., td.Distribution], param_dict: nn.ParameterDict, **kwargs):
        """
        Args:
            distribution_class: the class of the distribution that will be instantiated, e.g. torch.distributions.Normal
            param_dict: the dictionary of parameters to be passed to the distribution on instantiation, e.g. {"loc": ..., "scale": ...}
            **kwargs: Any other arguments to be passed to the distribution on instantiation (that aren't parameters), e.g. `validate_args=True`
        """
        super().__init__()
        self.param_dict = param_dict
        self.distribution_class = distribution_class
        self.kwargs = kwargs

    def forward(self) -> td.Distribution:
        """Return the instantiated `Distribution`"""
        return self.distribution_class(**dict(self.param_dict.items()), **self.kwargs)



import json
import os
from enum import Enum
from typing import Any, Counter, Dict, List, Optional, Set, Tuple

import fsspec
import numpy as np
import torch
from tensordict import TensorDict

from causica.datasets.interventional_data import CounterfactualData, InterventionData

CSUITE_DATASETS_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"


InterventionsWithEffects = List[Tuple[InterventionData, InterventionData, Set[str]]]
CounterfactualsWithEffects = List[Tuple[CounterfactualData, CounterfactualData, Set[str]]]

DTYPE_MAP = {"continuous": torch.float32, "categorical": torch.int32, "binary": torch.bool}


class DataEnum(Enum):
    TRAIN = "test.csv"
    TEST = "train.csv"
    INTERVENTIONS = "interventions.json"
    COUNTERFACTUALS = "counterfactuals.json"
    TRUE_ADJACENCY = "adj_matrix.csv"
    VARIABLES_JSON = "variables.json"


def get_csuite_path(dataset_path: str, dataset_name: str, data_enum: DataEnum) -> str:
    """
    Return the the absolute path to CSuite Data from the location, dataset name and type of data.

    Args:
        dataset_path: The root path to the CSuite Data e.g. `CSUITE_DATASETS_PATH`
        dataset_name: The name of the CSuite Data e.g. "csuite_linexp_2"
        data_enum: The type of dataset for which to return the path
    Return:
        The full path to the required CSuite Dataset
    """
    return os.path.join(dataset_path, dataset_name, data_enum.value)


def load_data(
    dataset_path: str,
    dataset_name: str,
    data_enum: DataEnum,
    variables_metadata: Optional[Dict[str, Any]] = None,
    **storage_options: Dict[str, Any],
):
    """
    Load the CSuite Data from the location, dataset name and type of data.

    Args:
        dataset_path: The root path to the CSuite Data e.g. `CSUITE_DATASETS_PATH`
        dataset_name: The name of the CSuite Data e.g. "csuite_linexp_2"
        data_enum: The type of dataset for which to return the path
        variables_metadata: Optional variables object (to save downloading it multiple times)
        **storage_options: Keyword args passed to `fsspec.open`
    Return:
        The downloaded CSuite data (the type depends on the data requested)
    """

    path_name = get_csuite_path(dataset_path, dataset_name, data_enum)

    if data_enum == DataEnum.TRUE_ADJACENCY:
        with fsspec.open(path_name, **storage_options) as f:
            return np.loadtxt(f, dtype=int, delimiter=",")

    if data_enum == DataEnum.VARIABLES_JSON:
        if variables_metadata is not None:
            raise ValueError("Variables metadata was supplied and requested")
        with fsspec.open(path_name, **storage_options) as f:
            return json.load(f)

    if variables_metadata is None:
        variables_metadata = load_data(dataset_path, dataset_name, data_enum=DataEnum.VARIABLES_JSON)

    with fsspec.open(path_name, **storage_options) as f:
        if data_enum in {DataEnum.TRAIN, DataEnum.TEST}:
            arr = np.loadtxt(f, delimiter=",")
            return tensordict_from_variables_metadata(arr, variables_metadata["variables"])
        elif data_enum == DataEnum.INTERVENTIONS:
            return _load_interventions(json_object=json.load(f), metadata=variables_metadata)
        elif data_enum == DataEnum.COUNTERFACTUALS:
            with fsspec.open(path_name) as f:
                return _load_counterfactuals(json_object=json.load(f), metadata=variables_metadata)
        else:
            raise RuntimeError("Fallthrough")


def _load_interventions(json_object: Dict[str, Any], metadata: Dict[str, Any]) -> InterventionsWithEffects:
    """
    Load the CSuite Interventional Datasets as a list of interventions/counterfactuals.

    Args:
        json_object: The .json file in CSuite loaded as a json object.
        metadata: Metadata of the dataset containing names and types.

    Returns:
        A list of interventions or counterfactuals and the nodes we want to observe for each
    """
    variables_list = metadata["variables"]

    intervened_column_to_group_name: Dict[int, str] = dict(
        zip(
            json_object["metadata"]["columns_to_nodes"],
            list(item["group_name"] for item in variables_list),
        )
    )

    interventions_list = []
    for environment in json_object["environments"]:
        conditioning_idxs = environment["conditioning_idxs"]
        if conditioning_idxs is None:
            condition_nodes = []
        else:
            condition_nodes = [intervened_column_to_group_name[idx] for idx in environment["conditioning_idxs"]]

        intervention_nodes = [intervened_column_to_group_name[idx] for idx in environment["intervention_idxs"]]

        intervention = _to_intervention(
            np.array(environment["test_data"]),
            intervention_nodes=intervention_nodes,
            condition_nodes=condition_nodes,
            variables_list=variables_list,
        )
        # if the json has reference data create another intervention dataclass
        if (reference_data := environment["reference_data"]) is None:
            raise RuntimeError()
        reference = _to_intervention(
            np.array(reference_data),
            intervention_nodes=intervention_nodes,
            condition_nodes=condition_nodes,
            variables_list=variables_list,
        )

        # store the nodes we're interested in observing
        effect_nodes = set(intervened_column_to_group_name[idx] for idx in environment["effect_idxs"])
        # default to all nodes
        if not effect_nodes:
            effect_nodes = set(intervention.sampled_nodes)
        interventions_list.append((intervention, reference, effect_nodes))
    return interventions_list


def _load_counterfactuals(json_object: Dict[str, Any], metadata: Dict[str, Any]) -> CounterfactualsWithEffects:
    """
    Load the CSuite Interventional Datasets as a list of interventions/counterfactuals.

    Args:
        json_object: The .json file in CSuite loaded as a json object.
        metadata: Metadata of the dataset containing names and types.

    Returns:
        A list of interventions or counterfactuals and the nodes we want to observe for each
    """
    variables_list = metadata["variables"]

    intervened_column_to_group_name: Dict[int, str] = dict(
        zip(
            json_object["metadata"]["columns_to_nodes"],
            list(item["group_name"] for item in variables_list),
        )
    )

    cf_list = []
    for environment in json_object["environments"]:
        factual_data = np.array(environment["conditioning_values"])
        intervention_nodes = [intervened_column_to_group_name[idx] for idx in environment["intervention_idxs"]]
        intervention = _to_counterfactual(
            np.array(environment["test_data"]),
            factual_data,
            intervention_nodes=intervention_nodes,
            variables_list=variables_list,
        )
        # if the json has reference data create another intervention dataclass
        if (reference_data := environment["reference_data"]) is None:
            raise RuntimeError()
        reference = _to_counterfactual(
            np.array(reference_data), factual_data, intervention_nodes=intervention_nodes, variables_list=variables_list
        )

        # store the nodes we're interested in observing
        effect_nodes = set(intervened_column_to_group_name[idx] for idx in environment["effect_idxs"])
        # default to all nodes
        if not effect_nodes:
            effect_nodes = set(intervention.sampled_nodes)
        cf_list.append((intervention, reference, effect_nodes))
    return cf_list


def _to_intervention(
    data: np.ndarray, intervention_nodes: List[str], condition_nodes: List[str], variables_list: List[Dict[str, Any]]
) -> InterventionData:
    """Create an `InterventionData` object from the data within the json file."""
    interv_data = tensordict_from_variables_metadata(data, variables_list=variables_list)
    # all the intervention values in the dataset should be the same, so we use the first row
    first_row = interv_data[0]
    assert all(torch.allclose(interv_data[node_name], first_row[node_name]) for node_name in intervention_nodes)
    intervention_values = TensorDict(
        {node_name: first_row[node_name] for node_name in intervention_nodes}, batch_size=tuple()
    )

    condition_values = TensorDict(
        {node_name: first_row[node_name] for node_name in condition_nodes}, batch_size=tuple()
    )
    return InterventionData(
        intervention_values=intervention_values,
        intervention_data=interv_data,
        condition_values=condition_values,
    )


def _to_counterfactual(
    data: np.ndarray, base_data: np.ndarray, intervention_nodes: List[str], variables_list: List[Dict[str, Any]]
) -> CounterfactualData:
    """Create an `CounterfactualData` object from the data within the json file."""
    interv_data = tensordict_from_variables_metadata(data, variables_list=variables_list)
    # all the intervention values in the dataset should be the same, so we use the first row
    first_row = interv_data[0]
    assert all(torch.allclose(interv_data[node_name], first_row[node_name]) for node_name in intervention_nodes)
    intervention_values = TensorDict(
        {node_name: first_row[node_name] for node_name in intervention_nodes}, batch_size=tuple()
    )

    return CounterfactualData(
        intervention_values=intervention_values,
        counterfactual_data=interv_data,
        factual_data=tensordict_from_variables_metadata(base_data, variables_list=variables_list),
    )


def get_categorical_sizes(variables_list: List[Dict[str, Any]]) -> Dict[str, int]:
    categorical_sizes = {}
    for item in variables_list:
        if item["type"] == "categorical":
            upper = item.get("upper")
            lower = item.get("lower")
            if upper is not None and lower is not None:
                categorical_sizes[item["group_name"]] = item["upper"] - item["lower"] + 1
            else:
                assert upper is None and lower is None, "Please specify either both limits or neither"
                categorical_sizes[item["group_name"]] = -1
    return categorical_sizes


def tensordict_from_variables_metadata(data: np.ndarray, variables_list: List[Dict[str, Any]]) -> TensorDict:
    """
    Convert a 2D numpy array and variables information into a `TensorDict`.
    """
    assert data.ndim == 2, "Numpy loading only supported for 2d data"
    batch_size = data.shape[0]

    # guaranteed to be ordered correctly in python 3.7+ https://docs.python.org/3/library/collections.html#collections.Counter
    sizes = Counter(d["group_name"] for d in variables_list)  # get the dimensions of each key from the variables
    assert sum(sizes.values()) == data.shape[1], "Variable sizes do not match data shape"

    dtypes = {item["group_name"]: DTYPE_MAP[item["type"]] for item in variables_list}

    # slice the numpy array and assign the slices to the values of keys in the dictionary
    d = TensorDict({}, batch_size=batch_size)
    curr_idx = 0
    for key, length in sizes.items():
        d[key] = torch.Tensor(data[:, curr_idx : curr_idx + length]).to(dtype=dtypes[key])
        curr_idx += length

    return d



import functools
import operator
from typing import Callable, Dict, List

import torch
import torch.distributions as td
from tensordict import TensorDict

from causica.distributions import NoiseAccessibleDistribution
from causica.functional_relationships import FunctionalRelationships, create_do_functional_relationship
from causica.sem.structural_equation_model import SEM


class DistributionParametersSEM(SEM):
    """
    A Structural Equation Model where the functional forward pass generates the parameters of the distributions to be sampled.

    This is more general than the simple additive case.
    """

    arg_constraints: Dict = {}

    def __init__(
        self,
        graph: torch.Tensor,
        node_names: List[str],
        noise_dist: Dict[str, Callable[[torch.Tensor], NoiseAccessibleDistribution]],
        func: FunctionalRelationships,
    ):
        """
        Args:
            graph: Adjacency Matrix
            node_names: The names of each node in the graph
            noise_dist: A dictionary of callables for generating distributions from predictions
            func: The functional relationship that can map from the tensor representation to the to itself.
        """
        super().__init__(graph=graph, node_names=node_names)
        self.noise_dist = noise_dist
        self.func = func
        set_node_names = set(node_names)
        assert set_node_names == noise_dist.keys() == set_node_names

        dist_dict = self._default_init()
        for dist in dist_dict.values():
            assert len(dist.event_shape) == 1, "All node distributions must have vector (rank-1) events"

    def log_prob(self, value: TensorDict) -> torch.Tensor:
        """
        Compute the log prob of the observations

        Args:
            value: a dictionary of sample_shape + batch_shape + [variable_dim] shaped tensors
        Return:
            A tensor of shape sample_shape + batch_shape of the log probability of each event
        """
        dist_params = self.func(value, self.graph)
        cur_log_list = [dist(dist_params[key]).log_prob(value[key]) for key, dist in self.noise_dist.items()]
        return functools.reduce(operator.iadd, cur_log_list)

    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape sample_shape + batch_shape + event_shape
        """
        X = noise.clone().zero_()
        for _ in self.node_names:
            forward = self.func(X, self.graph)
            X.update_({key: self.noise_dist[key](val).noise_to_sample(noise[key]) for key, val in forward.items()})
        return X

    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape sample_shape + batch_shape + event_shape
        Return:
            Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        """
        forward = self.func(sample, self.graph)
        return TensorDict(
            {key: self.noise_dist[key](val).sample_to_noise(sample[key]) for key, val in forward.items()},
            batch_size=sample.batch_size,
        )

    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape sample_shape + batch_shape + event_shape
        """
        dist_dict = self._default_init()
        return TensorDict({key: dist.sample(sample_shape) for key, dist in dist_dict.items()}, batch_size=sample_shape)

    def do(self, interventions: TensorDict) -> "DistributionParametersSEM":
        """Return the SEM associated with the interventions"""
        do_func, do_graphs = create_do_functional_relationship(
            interventions=interventions, func=self.func, graph=self.graph
        )
        unintervened_node_names = [name for name in self.node_names if name not in interventions.keys()]
        intervened_noise_dist = {name: self.noise_dist[name] for name in unintervened_node_names}
        return type(self)(
            graph=do_graphs, node_names=unintervened_node_names, noise_dist=intervened_noise_dist, func=do_func
        )

    def _default_init(self) -> Dict[str, td.Distribution]:
        """
        Initialise the distributions with a default value of zero.

        This is useful for checking properties of the distributions and sampling pure noise

        Using zero to initialise could be a problem for some distributions, but isn't at the minute.
        """
        return {
            key: dist_func(torch.zeros(self.func.variables[key], device=self.graph.device))
            for key, dist_func in self.noise_dist.items()
        }



from typing import Any

import pytorch_lightning as pl
import torch
from pytorch_lightning.utilities.types import STEP_OUTPUT

from causica.training.auglag import AugLagLossCalculator, AugLagLR


class AuglagLRCallback(pl.Callback):
    """Wrapper Class to make the Auglag Learning Rate Scheduler compatible with Pytorch Lightning"""

    def __init__(self, scheduler: AugLagLR):
        """
        Args:
            scheduler: The auglag learning rate scheduler to wrap
        """
        self.scheduler = scheduler

    def on_train_batch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        _ = trainer
        _ = batch
        _ = batch_idx
        assert isinstance(outputs, dict)
        optimizer = pl_module.optimizers()
        assert isinstance(optimizer, torch.optim.Optimizer)
        auglag_loss: AugLagLossCalculator = pl_module.auglag_loss  # type: ignore

        self.scheduler.step(
            optimizer=optimizer,
            loss=auglag_loss,
            loss_value=outputs["loss"].item(),
            lagrangian_penalty=outputs["constraint"].item(),
        )



import math
from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, List

import torch

from causica.datasets.csuite_data import CounterfactualsWithEffects, InterventionsWithEffects
from causica.sem.structural_equation_model import SEM, ate, ite


def eval_intervention_likelihoods(sems: List[SEM], interventions: InterventionsWithEffects) -> torch.Tensor:
    """
    Calculate the average log-prob of interventional data.

    Specifically we calculate 𝔼_sample[log(𝔼_G[p(sample | G)])]

    Args:
        sems: An iterable of SEMS to evaluate the interventional log prob of
        interventions (InterventionsWithEffects): True interventional data to use for evaluation.

    Returns:
        float: Log-likelihood of the interventional data.
    """
    interventions_iterator = chain.from_iterable((a, b) for a, b, _ in interventions)
    total_log_sum_exp_per_int = []  # this will end up being length number of interventions
    num_samples = 0  # keep track of the total number of samples
    for intervention in interventions_iterator:
        inner_log_probs = [
            sem.do(interventions=intervention.intervention_values).log_prob(intervention.intervention_data)
            for sem in sems
        ]
        # calculate log(Σ_G p(sample | G)) for each sample
        log_prob = list_logsumexp(inner_log_probs)  # batch_size
        assert len(log_prob.shape) == 1
        num_samples += log_prob.shape[-1]
        total_log_sum_exp_per_int.append(log_prob.sum())

    # Σ_sample log(Σ_G p(sample | G))
    total_log_prob = torch.stack(total_log_sum_exp_per_int, dim=0).sum(dim=0)
    # take the average over the datapoints and divide this by the number of graphs (inside the log)
    return total_log_prob / num_samples - math.log(len(sems))


def eval_ate_rmse(
    sems: Iterable[SEM], interventions: InterventionsWithEffects, samples_per_graph: int = 1000
) -> torch.Tensor:
    """Evaluate the ATEs of a model

    Args:
        sems: An iterable of structural equation models to evaluate the ATE RMSE of
        interventions (InterventionsWithEffects): True interventional data to use for evaluation.
        samples_per_graph (int, optional): Number of samples to draw per graph to calculate the ATE. Defaults to 1000.

    Returns:
        float: RMSE of the ATEs.
    """

    ate_rmses = []
    for (intervention_a, intervention_b, effects) in interventions:

        true_ates = {
            effect: intervention_a.intervention_data[effect].mean() - intervention_b.intervention_data[effect].mean()
            for effect in effects
        }

        # generate samples from the intervened distribution and the base distribution
        ates_per_graph: Dict[str, List[torch.Tensor]] = defaultdict(list)
        for sem in sems:
            graph_ates = ate(
                sem, intervention_a.intervention_values, intervention_b.intervention_values, effects, samples_per_graph
            )
            for key in effects:
                ates_per_graph[key].append(graph_ates[key].detach())

        generated_ates = {k: list_mean(v) for k, v in ates_per_graph.items()}

        ate_rmse = torch.sqrt(list_mean([torch.sum((generated_ates[key] - true_ates[key]) ** 2) for key in effects]))
        ate_rmses.append(ate_rmse)
    return list_mean(ate_rmses)


def eval_ite_rmse(sems: Iterable[SEM], counterfactuals: CounterfactualsWithEffects) -> torch.Tensor:
    """Evaluate the ITEs of a model.

    Args:
        sems: An iterable of structural equation models to evaluate the ITE RMSE of
        counterfactuals (CounterfactualsWithEffects): Data of true counterfactuals to use for evaluation.

    Returns:
        float: RMSE of the ITEs.
    """
    ite_rmses = []
    for (intervention_a, intervention_b, effects) in counterfactuals:
        for key, a_val in intervention_a.factual_data.items():
            b_val = intervention_b.factual_data[key]
            assert torch.allclose(a_val, b_val), "Base data must be the same for ITEs"
        true_ites = {
            effect: intervention_a.counterfactual_data[effect] - intervention_b.counterfactual_data[effect]
            for effect in effects
        }

        factual_data_dict = intervention_a.factual_data

        # generate samples from the intervened distribution and the base distribution
        per_graph_ites: Dict[str, List[torch.Tensor]] = defaultdict(list)
        for sem in sems:
            sem_ites = ite(
                sem, factual_data_dict, intervention_a.intervention_values, intervention_b.intervention_values, effects
            )

            for key in effects:
                per_graph_ites[key].append(sem_ites[key].detach())

        generated_ites = {k: list_mean(v) for k, v in per_graph_ites.items()}

        ite_rmse = torch.sqrt(
            list_mean([torch.sum((generated_ites[key] - true_ites[key]) ** 2, -1) for key in effects])
        )
        ite_rmses.append(ite_rmse)

    return torch.mean(list_mean(ite_rmses))


def list_mean(list_of_tensors: List[torch.Tensor]) -> torch.Tensor:
    """Take the mean of a list of torch tensors, they must all have the same shape"""
    return torch.stack(list_of_tensors, dim=0).mean(dim=0)


def list_logsumexp(list_of_tensors: List[torch.Tensor]) -> torch.Tensor:
    """Take the logsumexp of a list of torch tensors, they must all have the same shape"""
    return torch.logsumexp(torch.stack(list_of_tensors, dim=0), dim=0)



from .adjacency_distributions import AdjacencyDistribution
from .constrained_adjacency_distributions import ConstrainedAdjacencyDistribution, constrained_adjacency
from .enco import ENCOAdjacencyDistribution
from .gibbs_dag_prior import ExpertGraphContainer, GibbsDAGPrior
from .three_way import ThreeWayAdjacencyDistribution



import torch
import torch.distributions as td

from causica.distributions.noise_accessible.noise_accessible import NoiseAccessibleDistribution


class NoiseAccessibleMultivariateNormal(td.MultivariateNormal, NoiseAccessibleDistribution):
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return samples - self.loc

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return noise + self.loc



import abc
import json
from functools import partial
from typing import Any, Dict, OrderedDict

import fsspec
import numpy as np
import pytorch_lightning as pl
import torch
from tensordict import TensorDict
from torch.utils.data import DataLoader

from causica.datasets.csuite_data import (
    DataEnum,
    get_categorical_sizes,
    get_csuite_path,
    tensordict_from_variables_metadata,
)
from causica.datasets.tensordict_utils import convert_one_hot, tensordict_shapes


class DECIDataModule(abc.ABC, pl.LightningDataModule):
    """Data Module for use with the `DECILightningModule`.

    Requires variable shapes to be available before the `setup` step.
    """

    @abc.abstractmethod
    def get_variable_shapes(self) -> OrderedDict[str, torch.Size]:
        """Return the shapes for each variable name."""
        pass

    @abc.abstractmethod
    def get_variable_types(self) -> OrderedDict[str, str]:
        """Return the types for each variable name."""


class VariableSpecDataModule(DECIDataModule):
    """Loads training and test data from fully specified paths for `variables.json` formatted data.

    This format assumes the `variables.json` specified all metadata, and that the corresponding CSV files do not have
    any header rows.

    Note:
        Uses `fsspec` to load the data in the paths. To load from a cloud storage location, make sure the relevant
        `fsspec` plugin is available and provide its corresponding scheme. Provide **storage_options for authentication.
        E.g., to load from an Azure storage account, install `adlfs` and use `az://container@storage_account/path`.

    Args:
        train_path: Path to training CSV data.
        test_path: Path to test CSV data.
        variable_path: Path to `variables.json` file.
        batch_size: Batch size for training and test data.
        dataset_name: Name for dataset.
        storage_options: Storage options forwarded to `fsspec` when loading files.
    """

    def __init__(
        self,
        train_path: str,
        test_path: str,
        variables_path: str,
        batch_size: int = 128,
        dataset_name: str = "anonymous_dataset",
        **storage_options: Dict[str, Any]
    ):
        super().__init__()
        self.batch_size = batch_size
        self.dataset_name = dataset_name
        self.train_path = train_path
        self.test_path = test_path
        self.variables_path = variables_path
        self.batch_size = batch_size
        self.storage_options = storage_options

    def get_variable_shapes(self) -> OrderedDict[str, torch.Size]:
        if self._variable_shapes is None:
            raise ValueError("Tried to get variable group shapes before data was downloaded.")
        return self._variable_shapes

    def get_variable_types(self) -> OrderedDict[str, str]:
        if self._variable_types is None:
            raise ValueError("Tried to get variable group types before data was downloaded.")
        return self._variable_types

    def prepare_data(self):
        fsspec_open = partial(fsspec.open, mode="r", encoding="utf-8", **self.storage_options)
        with fsspec_open(self.variables_path) as f:
            variables_metadata = json.load(f)

        categorical_sizes = get_categorical_sizes(variables_list=variables_metadata["variables"])
        with fsspec_open(self.train_path) as f:
            arr = np.loadtxt(f, delimiter=",")
            dataset_train: TensorDict = tensordict_from_variables_metadata(arr, variables_metadata["variables"])
            self.dataset_train = convert_one_hot(dataset_train, one_hot_sizes=categorical_sizes)

        with fsspec_open(self.test_path) as f:
            arr = np.loadtxt(f, delimiter=",")
            self.dataset_test: TensorDict = tensordict_from_variables_metadata(arr, variables_metadata["variables"])
            self.dataset_test = convert_one_hot(dataset_train, one_hot_sizes=categorical_sizes)

        assert set(self.dataset_train.keys()) == set(
            self.dataset_test.keys()
        ), "the groups_names for the training and test data must match"
        self._variable_shapes = tensordict_shapes(self.dataset_train)
        self._variable_types = {var["group_name"]: var["type"] for var in variables_metadata["variables"]}

    def train_dataloader(self):
        return DataLoader(
            dataset=self.dataset_train,
            collate_fn=lambda x: x,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.dataset_test,
            collate_fn=lambda x: x,
            batch_size=self.batch_size,
            shuffle=False,
            drop_last=False,
        )


class CSuiteDataModule(VariableSpecDataModule):
    """CSuite data module for loading by the datasets name.

    Args:
        dataset_name: Name of the dataset to load.
        batch_size: Batch size for all datasets.
        dataset_path: Path to CSuite dataset mirror.
    """

    DEFAULT_CSUITE_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"

    def __init__(
        self,
        dataset_name: str,
        batch_size: int = 128,
        dataset_path: str = DEFAULT_CSUITE_PATH,
    ):
        super().__init__(
            train_path=get_csuite_path(dataset_path, dataset_name, DataEnum.TRAIN),
            test_path=get_csuite_path(dataset_path, dataset_name, DataEnum.TEST),
            variables_path=get_csuite_path(dataset_path, dataset_name, DataEnum.VARIABLES_JSON),
            batch_size=batch_size,
            dataset_name=dataset_name,
        )



from .adjacency import (
    AdjacencyDistribution,
    ConstrainedAdjacencyDistribution,
    ENCOAdjacencyDistribution,
    ExpertGraphContainer,
    GibbsDAGPrior,
    ThreeWayAdjacencyDistribution,
    constrained_adjacency,
)
from .noise_accessible import (
    NoiseAccessible,
    NoiseAccessibleBernoulli,
    NoiseAccessibleCategorical,
    NoiseAccessibleDistribution,
    NoiseAccessibleIndependent,
    NoiseAccessibleMultivariateNormal,
)
from .parametrized_distribution import ParametrizedDistribution
from .splines import SplineDistribution, SplineParamListType, create_spline_dist_params



import fsspec
import numpy as np
import torch

from causica.distributions.adjacency import ExpertGraphContainer


class LoadedExpertGraphContainer(ExpertGraphContainer):
    """Loads an expert graph container from a file path."""

    def __init__(self, graph_path: str, confidence: float, scale: float):
        """
        Args:
            graph_path: fsspec compatible path to graph saved as a numpy array, with np.nan indicating undecided values.
            confidence: See ExpertGraphContainer
            scale: See ExpertGraphContainer
        """
        with fsspec.open(graph_path, "rb", encoding="utf-8") as f:
            prior_adj_matrix = np.load(f)
        prior_mask = ~np.isnan(prior_adj_matrix)
        super().__init__(
            dag=torch.tensor(prior_adj_matrix), mask=torch.tensor(prior_mask), confidence=confidence, scale=scale
        )



"""
An example script showing how to train a DECI model on CSuite Data.

This demonstrates how to assemble the various components of the library and how to perform training.

This logs MLFlow logs to the environment variable location if it is set https://mlflow.org/docs/latest/tracking.html#where-runs-are-recorded
"""
import argparse
import logging
import pathlib
from dataclasses import dataclass

import fsspec
import mlflow
import pytorch_lightning as pl
import torch
from dataclasses_json import DataClassJsonMixin
from tensordict import TensorDict
from torch.nn import Parameter, ParameterDict
from torch.optim import Adam
from torch.utils.data import DataLoader

from causica.datasets.csuite_data import DataEnum, get_categorical_sizes, load_data
from causica.datasets.tensordict_utils import convert_one_hot, tensordict_shapes
from causica.distributions import (
    AdjacencyDistribution,
    ENCOAdjacencyDistribution,
    GibbsDAGPrior,
    ParametrizedDistribution,
)
from causica.functional_relationships import ICGNN
from causica.graph.dag_constraint import calculate_dagness
from causica.sem.distribution_parameters_sem import DistributionParametersSEM
from causica.training.auglag import AugLagLossCalculator, AugLagLR, AugLagLRConfig
from causica.training.trainable_container import NoiseDist, TrainableContainer, create_noise_dists
from causica.training.training_callbacks import AverageMetricTracker

logging.basicConfig(level=logging.INFO)

BEST_MODEL = "best_model.pt"
LAST_MODEL = "last_model.pt"


DATASET_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"


# Data class for config
@dataclass(frozen=True)
class TrainingConfig(DataClassJsonMixin):  # using mixin to prevent mypy errors
    noise_dist: NoiseDist = NoiseDist.GAUSSIAN
    batch_size: int = 128
    max_epoch: int = 1000
    gumbel_temp: float = 0.25
    averaging_period: int = 10
    prior_sparsity_lambda: float = 0.05
    # Auglag parameters
    init_alpha: float = 0.0
    init_rho: float = 1.0


@dataclass(frozen=True)
class ICGNNConfig(DataClassJsonMixin):
    embedding_size: int = 32
    out_dim_g: int = 32
    norm_layer: bool = True
    res_connection: bool = True


def train(
    dataset: str, seed: int, auglag_config: AugLagLRConfig, icgnn_config: ICGNNConfig, training_config: TrainingConfig
) -> TrainableContainer:
    """Run training on CSuite with a noise distribution given by `noise_dist_enum`."""
    pl.seed_everything(seed)  # set the random seed
    device = "cuda" if torch.cuda.is_available() else "cpu"

    variables_metadata = load_data(DATASET_PATH, dataset, DataEnum.VARIABLES_JSON)
    dataset_train: TensorDict = load_data(DATASET_PATH, dataset, DataEnum.TRAIN, variables_metadata)

    dataset_train = convert_one_hot(
        dataset_train, one_hot_sizes=get_categorical_sizes(variables_list=variables_metadata["variables"])
    )

    dataloader_train = DataLoader(
        dataset=dataset_train,
        collate_fn=lambda x: x,
        batch_size=training_config.batch_size,
        shuffle=True,
        drop_last=False,
    )

    num_nodes = len(dataset_train.keys())

    # Define vardist params
    vardist_logits_orient = Parameter(torch.zeros(int(num_nodes * (num_nodes - 1) / 2)), requires_grad=True)
    vardist_logits_exist = Parameter(torch.zeros(num_nodes, num_nodes), requires_grad=True)

    param_vardist = ParametrizedDistribution(
        ENCOAdjacencyDistribution,
        ParameterDict(dict(logits_orient=vardist_logits_orient, logits_exist=vardist_logits_exist)),
    )

    # Define ICGNN
    icgnn = ICGNN(
        variables=tensordict_shapes(dataset_train),
        embedding_size=icgnn_config.embedding_size,
        out_dim_g=icgnn_config.out_dim_g,
        norm_layer=torch.nn.LayerNorm if icgnn_config.norm_layer else None,
        res_connection=icgnn_config.res_connection,
    )
    # create the noise distributions
    variables = load_data(DATASET_PATH, dataset, DataEnum.VARIABLES_JSON)["variables"]
    types_dict = {var["group_name"]: var["type"] for var in variables}
    noise_dist_funcs, param_noise_dist = create_noise_dists(
        tensordict_shapes(dataset_train), types_dict, training_config.noise_dist
    )

    # Create container
    container = TrainableContainer(
        dataset_name=dataset,
        icgnn=icgnn,
        vardist=param_vardist,
        noise_dist_params=param_noise_dist,
        noise_dist_type=training_config.noise_dist,
    )
    container.to(device)

    # Create prior
    prior = GibbsDAGPrior(num_nodes=num_nodes, sparsity_lambda=training_config.prior_sparsity_lambda)

    # Define Optimizer
    # @TODO: Clean this up
    parameter_list = [
        {
            "params": module if isinstance(module, torch.nn.Parameter) else module.parameters(),
            "lr": auglag_config.lr_init_dict[name],
            "name": name,
        }
        for name, module in container.named_children()
    ]
    optimizer = Adam(parameter_list)

    # Define training helper classes
    scheduler = AugLagLR(config=auglag_config)
    avg_metric_tracker = AverageMetricTracker(averaging_period=training_config.averaging_period)
    auglag_loss = AugLagLossCalculator(init_alpha=training_config.init_alpha, init_rho=training_config.init_rho)

    assert len(dataset_train.batch_size) == 1, "Only 1D batch size is supported"
    num_samples = len(dataset_train)
    step_counter = 0
    for epoch in range(training_config.max_epoch):
        for batch in dataloader_train:
            batch = batch.apply(lambda t: t.to(dtype=torch.float32, device=device))
            optimizer.zero_grad()
            # sample graph
            vardist = param_vardist.forward()
            assert isinstance(vardist, AdjacencyDistribution)
            cur_graph = vardist.relaxed_sample(torch.Size([]), temperature=training_config.gumbel_temp)  # soft sample
            # Create SEM
            sem = DistributionParametersSEM(
                graph=cur_graph,
                node_names=dataset_train.keys(),
                noise_dist=noise_dist_funcs,
                func=container.icgnn,
            )

            batch_log_prob = sem.log_prob(batch).mean()  # average over graphs

            objective = (-vardist.entropy() - prior.log_prob(cur_graph)) / num_samples - batch_log_prob
            constraint = calculate_dagness(cur_graph)
            weighted_constraint = constraint / num_samples

            loss = auglag_loss(objective, weighted_constraint)  # we should use weighted constraint for the loss

            loss.backward()
            optimizer.step()
            scheduler.step(
                optimizer=optimizer,
                loss=auglag_loss,
                loss_value=loss.item(),
                lagrangian_penalty=constraint.item(),
            )
            if avg_metric_tracker.step(-batch_log_prob.item()):
                print(f"Saving model: {avg_metric_tracker.min_value}")
                torch.save({"container": container}, BEST_MODEL)

            # log metrics
            print(
                f"epoch:{epoch} loss:{loss.item():.5g} nll:{-batch_log_prob.detach().cpu().numpy():.5g} dagness:{constraint.item():.5f} "
                f"num_edges:{(cur_graph > 0.0).sum()} alpha:{auglag_loss.alpha:.5g} rho:{auglag_loss.rho:.5g} "
                f"step:{scheduler.outer_opt_counter}/{scheduler.step_counter} num_lr_updates:{scheduler.num_lr_updates}"
            )
            step_counter += 1

    # save the final model
    torch.save({"container": container}, LAST_MODEL)
    return container


def main():
    this_dir = pathlib.Path().resolve()
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--dataset", "-d", default="csuite_nonlingauss")
    parser.add_argument("--seed", "-s", default=123, type=int)

    parser.add_argument(
        "--config-auglag", type=str, default=f"{this_dir}/auglag_default.json", dest="auglag_config_path"
    )
    parser.add_argument("--config-icgnn", type=str, default=f"{this_dir}/icgnn_default.json", dest="icgnn_config_path")
    parser.add_argument(
        "--config-training", type=str, default=f"{this_dir}/training_default.json", dest="training_config_path"
    )

    args = parser.parse_args()

    # Load configs
    with fsspec.open(args.auglag_config_path, "r", encoding="utf-8") as f:
        auglag_config = AugLagLRConfig.from_json(f.read())

    with fsspec.open(args.icgnn_config_path, "r", encoding="utf-8") as f:
        icgnn_config = ICGNNConfig.from_json(f.read())

    with fsspec.open(args.training_config_path, "r", encoding="utf-8") as f:
        training_config = TrainingConfig.from_json(f.read())

    with mlflow.start_run(run_name=f"csuite_{args.dataset}", tags={"dataset": args.dataset, "seed": args.seed}) as run:
        if mlflow.get_tracking_uri().startswith("azureml://"):
            uri = f"https://ml.azure.com/experiments/id/{run.info.experiment_id}/runs/{run.info.run_id}"
            run_name = run.to_dictionary()["data"]["tags"].get("mlflow.runName", "")
            print(f"Started experiment {run_name}: {uri}")

        train(
            dataset=args.dataset,
            seed=args.seed,
            auglag_config=auglag_config,
            icgnn_config=icgnn_config,
            training_config=training_config,
        )


if __name__ == "__main__":
    main()



"""
Implementation of the Piecewise Rational Quadratic Transform

This is pretty much a copy-paste of
    https://github.com/tonyduan/normalizing-flows/blob/master/nf/utils.py

We should consider using the Pyro implementation.
"""
from typing import Tuple

import numpy as np
import torch
import torch.distributions as td
from torch.nn import functional as F

DEFAULT_MIN_BIN_WIDTH = 1e-3
DEFAULT_MIN_BIN_HEIGHT = 1e-3
DEFAULT_MIN_DERIVATIVE = 1e-3


class PiecewiseRationalQuadraticTransform(td.Transform):
    """
    Layer that implements a spline-cdf (https://arxiv.org/abs/1906.04032) transformation.

    All dimensions of x are treated as independent, no coupling is used. This is needed
    to ensure invertibility in our additive noise SEM.

    This is pretty much a copy-paste of
        https://github.com/tonyduan/normalizing-flows/blob/master/nf/utils.py
    """

    bijective = True

    def __init__(self, knot_locations: torch.Tensor, derivatives: torch.Tensor, tail_bound: float = 3.0):
        """
        Args:
            knot_locations: the x, y points of the knots,  shape [dim, num_bins, 2]
            derivatives: the derivatives at the knots, shape [dim, num_bins - 1]
            tail_bound: distance of edgemost bins relative to 0,
        """
        super().__init__()
        assert knot_locations.ndim == 3, "Only two dimensional params are supported"
        assert knot_locations.shape[-1] == 2, "Knot locations are 2-d"
        self.dim, self.num_bins, *_ = knot_locations.shape

        assert derivatives.shape == (self.dim, self.num_bins - 1)

        self.tail_bound = tail_bound
        self._event_dim = 0

        self.knot_locations = knot_locations
        self.derivatives = derivatives

    @property
    def event_dim(self):
        return self._event_dim

    def _piecewise_cdf(self, inputs: torch.Tensor, inverse: bool = False) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Evaluate the Cumulative Density function at `inputs`

        Args:
            inputs: the positions at which to evaluate the cdf shape batch_shape + (input_dim)
            inverse: whether this is forwards or backwards transform
        Returns:
            input_evaluations and absolute log determinants, a tuple of tensors of shape batch_shape + (input_dim)
        """
        assert len(inputs.shape) == 2  # TODO(JJ) accept 1d inputs
        batch_shape = inputs.shape[:-1]
        # shape batch_shape + (dim, 3 * num_bins - 1)
        expanded_knot_locations = self.knot_locations[None, ...].expand(*batch_shape, -1, -1, -1)

        unnormalized_derivatives = F.pad(
            self.derivatives[None, ...].expand(*batch_shape, -1, -1),
            pad=(1, 1),
            value=np.log(np.expm1(1 - DEFAULT_MIN_DERIVATIVE)),
        )  # shape batch_shape + (dim, num_bins + 1)

        inside_intvl_mask = torch.abs(inputs) <= self.tail_bound
        outputs = inputs.clone().detach()  # shape batch_shape + (input_dim)
        outputs[inside_intvl_mask] = 0

        logabsdet = torch.zeros_like(inputs)  # shape batch_shape + (input_dim)

        if inside_intvl_mask.any():
            outputs[inside_intvl_mask], logabsdet[inside_intvl_mask] = rational_quadratic_spline(
                inputs=inputs[inside_intvl_mask],
                unnormalized_knot_locations=expanded_knot_locations[inside_intvl_mask, :, :],
                unnormalized_derivatives=unnormalized_derivatives[inside_intvl_mask, :],
                inverse=inverse,
                tail_bound=self.tail_bound,
            )
        return outputs, logabsdet

    @td.constraints.dependent_property(is_discrete=False)
    def domain(self):
        if self.event_dim == 0:
            return td.constraints.real
        return td.constraints.independent(td.constraints.real, self.event_dim)

    @td.constraints.dependent_property(is_discrete=False)
    def codomain(self):
        if self.event_dim == 0:
            return td.constraints.real
        return td.constraints.independent(td.constraints.real, self.event_dim)

    def _call(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: batch_shape + (input_dim)
        Returns:
            transformed_input batch_shape + (input_dim)
        """
        return self._piecewise_cdf(x, inverse=False)[0]

    def _inverse(self, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            y: batch_shape + (input_dim)
        Returns:
            transformed_input, batch_shape + (input_dim)
        """
        return self._piecewise_cdf(y, inverse=True)[0]

    def log_abs_det_jacobian(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            input: batch_shape + (input_dim)
        Returns:
            jacobian_log_determinant: batch_shape + (input_dim)
        """
        return self._piecewise_cdf(x, inverse=False)[1]


def rational_quadratic_spline(
    inputs: torch.Tensor,
    unnormalized_knot_locations: torch.Tensor,
    unnormalized_derivatives: torch.Tensor,
    inverse: bool = False,
    tail_bound: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Args:
        inputs: shape batch_shape + (input_dim)
        unnormalized_knot_locations: shape batch_shape + (dim, num_bins, 2)
        unnormalized_derivatives: shape batch_shape + (dim, num_bins + 1)
        inverse: whether this is the forward or backward pass
        tail_bound:
    Returns:
        the outputs and the absolute log determinant, a tuple of tensors of shape batch_shape + (input_dim)
    """

    if (torch.abs(inputs) > tail_bound).any():
        raise ValueError("Input outside domain")

    num_bins = unnormalized_knot_locations.shape[-2]

    if DEFAULT_MIN_BIN_WIDTH * num_bins > 1.0:
        raise ValueError("Minimal bin width too large for the number of bins")
    if DEFAULT_MIN_BIN_HEIGHT * num_bins > 1.0:
        raise ValueError("Minimal bin height too large for the number of bins")

    # create the bins
    derivatives = DEFAULT_MIN_DERIVATIVE + F.softplus(unnormalized_derivatives)
    # shape batch_shape + (dim)
    widths, cumwidths = normalize_bins(
        unnormalized_knot_locations[..., 0], tail_bound=tail_bound, min_bin_size=DEFAULT_MIN_BIN_WIDTH
    )
    heights, cumheights = normalize_bins(
        unnormalized_knot_locations[..., 1], tail_bound=tail_bound, min_bin_size=DEFAULT_MIN_BIN_HEIGHT
    )

    # place the input data within the bins
    bin_idx = torch.searchsorted(cumheights if inverse else cumwidths, inputs[..., None]) - 1

    input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

    input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
    input_bin_heights = heights.gather(-1, bin_idx)[..., 0]

    delta = heights / widths
    input_delta = delta.gather(-1, bin_idx)[..., 0]

    input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
    input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)
    input_derivatives_plus_one = input_derivatives_plus_one[..., 0]

    # here be dragons
    if inverse:
        a = (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        ) + input_bin_heights * (input_delta - input_derivatives)
        b = input_bin_heights * input_derivatives - (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        )
        c = -input_delta * (inputs - input_cumheights)

        discriminant = b * b - 4 * a * c
        assert (discriminant >= 0).all()

        theta = (2 * c) / (-b - torch.sqrt(discriminant))
        theta_one_minus_theta = theta * (1 - theta)

        outputs = theta * input_bin_widths + input_cumwidths

        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta) * theta_one_minus_theta
        )
        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * theta * theta
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - theta).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)
        return outputs, -logabsdet
    else:
        theta = (inputs - input_cumwidths) / input_bin_widths
        theta_one_minus_theta = theta * (1 - theta)

        numerator = input_bin_heights * (input_delta * theta * theta + input_derivatives * theta_one_minus_theta)
        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta) * theta_one_minus_theta
        )
        outputs = input_cumheights + numerator / denominator

        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * theta.pow(2)
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - theta).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)
        return outputs, logabsdet


def normalize_bins(
    bins: torch.Tensor, tail_bound: float, min_bin_size: float = 1e-3
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    For a tensor of unnormalized bin widths `bins`, create bins in [-tail_bound, tail_bound]

    Args:
        bins: shape (..., num_bins)
        tail_bound: the size of interval to project to
        min_bin_size: the minimum bin size
    Returns:
        Tuple of the normalized bin_widths and their locations shape (..., num_bins) and (..., num_bins + 1)
    """
    num_bins = bins.shape[-1]
    # shape batch_shape + (dim)
    norm_bins = min_bin_size + (1 - min_bin_size * num_bins) * F.softmax(bins, dim=-1)
    cum_norm_bins = F.pad(torch.cumsum(norm_bins, dim=-1), pad=(1, 0), mode="constant", value=0.0)
    cum_norm_bins = tail_bound * (2 * cum_norm_bins - 1)
    norm_bins = torch.diff(cum_norm_bins)
    return norm_bins, cum_norm_bins



import logging
from typing import Any, Dict, Optional, Sequence, Union

import fsspec
import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.trainer.states import TrainerFn
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter, ParameterDict

from causica.distributions import (
    AdjacencyDistribution,
    ENCOAdjacencyDistribution,
    ExpertGraphContainer,
    GibbsDAGPrior,
    ParametrizedDistribution,
    constrained_adjacency,
)
from causica.functional_relationships import ICGNN
from causica.graph.dag_constraint import calculate_dagness
from causica.lightning.callbacks import AuglagLRCallback
from causica.lightning.data_modules import DECIDataModule
from causica.sem.distribution_parameters_sem import DistributionParametersSEM
from causica.training.auglag import AugLagLossCalculator, AugLagLR, AugLagLRConfig
from causica.training.trainable_container import NoiseDist, TrainableContainer, create_noise_dists
from causica.training.training_callbacks import AverageMetricTracker

logging.basicConfig(level=logging.INFO)

DATASET_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"


class DECIModule(pl.LightningModule):
    """A PyTorch Lightning Module for running a DECI model."""

    def __init__(
        self,
        noise_dist: NoiseDist = NoiseDist.SPLINE,
        embedding_size: int = 32,
        out_dim_g: int = 32,
        norm_layer: bool = True,
        res_connection: bool = True,
        init_alpha: float = 0.0,
        init_rho: float = 1.0,
        prior_sparsity_lambda: float = 0.05,
        gumbel_temp: float = 0.25,
        auglag_config: Optional[AugLagLRConfig] = None,
        expert_graph_container: Optional[ExpertGraphContainer] = None,
        constraint_matrix_path: Optional[str] = None,
    ):
        super().__init__()
        self.auglag_config = auglag_config if auglag_config is not None else AugLagLRConfig()
        self.constraint_matrix_path = constraint_matrix_path
        self.constraint_matrix: Optional[torch.Tensor] = None
        self.embedding_size = embedding_size
        self.expert_graph_container = expert_graph_container
        self.gumbel_temp = gumbel_temp
        self.init_alpha = init_alpha
        self.init_rho = init_rho
        self.is_setup = False
        self.noise_dist = noise_dist
        self.norm_layer = norm_layer
        self.out_dim_g = out_dim_g
        self.prior_sparsity_lambda = prior_sparsity_lambda
        self.res_connection = res_connection

    def prepare_data(self) -> None:
        if self.constraint_matrix_path:
            with fsspec.open(self.constraint_matrix_path) as f:
                constraint_matrix = np.loadtxt(f, dtype=int, delimiter=",")
            self.constraint_matrix = torch.tensor(constraint_matrix)
        return super().prepare_data()

    def setup(self, stage: Optional[str] = None):
        if self.is_setup:
            return  # Already setup
        elif stage != TrainerFn.FITTING:
            raise ValueError(f"Model can only be setup during the {TrainerFn.FITTING} stage.")

        datamodule = self.trainer.datamodule  # type: ignore
        if not isinstance(datamodule, DECIDataModule):
            raise TypeError(
                f"Incompatible data module {datamodule}, requires a DECIDataModule but is " f"{type(datamodule).mro()}"
            )
        dataset_train = datamodule.dataset_train
        variable_group_shapes = datamodule.get_variable_shapes()

        self.node_names = list(variable_group_shapes)
        num_nodes = len(variable_group_shapes)
        vardist_logits_orient = Parameter(torch.zeros(int(num_nodes * (num_nodes - 1) / 2)), requires_grad=True)
        vardist_logits_exist = Parameter(torch.zeros(num_nodes, num_nodes), requires_grad=True)

        adjacency_distribution = ENCOAdjacencyDistribution
        if self.constraint_matrix is not None:
            adjacency_distribution = constrained_adjacency(
                adjacency_distribution, graph_constraint_matrix=self.constraint_matrix
            )

        param_vardist = ParametrizedDistribution(
            adjacency_distribution,
            ParameterDict({"logits_orient": vardist_logits_orient, "logits_exist": vardist_logits_exist}),
        )

        # Define ICGNN
        icgnn = ICGNN(
            variables=variable_group_shapes,
            embedding_size=self.embedding_size,
            out_dim_g=self.out_dim_g,
            norm_layer=None if self.norm_layer is False else torch.nn.LayerNorm,
            res_connection=self.res_connection,
        )

        self.num_samples = len(dataset_train)
        self.noise_dist_funcs, param_noise_dist = create_noise_dists(
            variable_group_shapes, datamodule.get_variable_types(), self.noise_dist
        )

        self.container = TrainableContainer(
            dataset_name=datamodule.dataset_name,
            icgnn=icgnn,
            vardist=param_vardist,
            noise_dist_params=param_noise_dist,
            noise_dist_type=self.noise_dist,
        )
        self.auglag_loss = AugLagLossCalculator(init_alpha=self.init_alpha, init_rho=self.init_rho)
        self.prior = GibbsDAGPrior(
            num_nodes=num_nodes,
            sparsity_lambda=self.prior_sparsity_lambda,
            expert_graph_container=self.expert_graph_container,
        )
        # TODO: Set a more reasonable averaging period
        self.average_batch_log_prob_tracker = AverageMetricTracker(averaging_period=10)
        self.is_setup = True

    def training_step(self, *args, **kwargs) -> STEP_OUTPUT:
        # sample graph
        _ = kwargs
        batch, *_ = args
        batch = batch.apply(lambda t: t.to(torch.float32))

        vardist = self.container.vardist()
        assert isinstance(vardist, AdjacencyDistribution)
        cur_graph = vardist.relaxed_sample(torch.Size([]), temperature=self.gumbel_temp)  # soft sample

        # Create SEM
        sem = DistributionParametersSEM(
            graph=cur_graph,
            node_names=self.node_names,
            noise_dist=self.noise_dist_funcs,
            func=self.container.icgnn,
        )

        batch_log_prob = sem.log_prob(batch).mean()
        objective = (-vardist.entropy() - self.prior.log_prob(cur_graph)) / self.num_samples - batch_log_prob
        constraint = calculate_dagness(cur_graph)
        self.average_batch_log_prob_tracker.step(batch_log_prob.item())
        step_output = {
            "loss": self.auglag_loss(objective, constraint / self.num_samples),
            "batch_log_prob": batch_log_prob,
            "constraint": constraint,
            "num_edges": (cur_graph > 0.0).count_nonzero(),
            "average_batch_log_prob": self.average_batch_log_prob_tracker.average,
        }
        self.log_dict({"alpha": self.auglag_loss.alpha, "rho": self.auglag_loss.rho, **step_output}, prog_bar=True)
        return step_output

    def configure_optimizers(self):
        # Define Optimizer
        parameter_list = [
            {
                "params": module if isinstance(module, Parameter) else module.parameters(),
                "lr": self.auglag_config.lr_init_dict[name],
                "name": name,
            }
            for name, module in self.container.named_children()
        ]
        return torch.optim.Adam(parameter_list)

    def configure_callbacks(self) -> Union[Sequence[pl.Callback], pl.Callback]:
        # TODO: Integrate with Lightning lr scheduler and set up in `configure_optimizers` instead
        auglag_callback = AuglagLRCallback(AugLagLR(config=self.auglag_config))
        return [auglag_callback]

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        checkpoint["container"] = self.container



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'examples': './examples',
 'examples.csuite_example': './examples/csuite_example'}

packages = \
['causica',
 'causica.datasets',
 'causica.distributions',
 'causica.distributions.adjacency',
 'causica.distributions.noise_accessible',
 'causica.distributions.splines',
 'causica.functional_relationships',
 'causica.graph',
 'causica.lightning',
 'causica.sem',
 'causica.training',
 'examples',
 'examples.csuite_example']

package_data = \
{'': ['*']}

install_requires = \
['azureml-mlflow>=1.46.0,<2.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'mlflow>=2.0.0,<3.0.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pytorch-lightning[extra]>=1.7.7,<2.0.0',
 'tensorboard>=2.9.0,<3.0.0',
 'tensordict>=0.0.2a0,<0.0.2',
 'types-PyYAML>=6.0.12.2,<7.0.0.0']

setup_kwargs = {
    'name': 'causica',
    'version': '0.1.0',
    'description': '',
    'long_description': '[![Causica CI Build](https://github.com/microsoft/causica/actions/workflows/ci-build.yml/badge.svg)](https://github.com/microsoft/causica/actions/workflows/ci-build.yml)\n\n# Causica\n\n## Overview\n \nCausal machine learning enables individuals and organizations to make better data-driven decisions. In particular, causal ML allows us to answer “what if” questions about the effect of potential actions on outcomes. \n \nCausal ML is a nascent area, we aim  to enable a **scalable**, **flexible**, **real-world applicable end-to-end** causal inference framework. In perticular, we bridge between causal discovery, causal inference, and deep learning to achieve the goal.  We aim to develop technology can automate causal decision-making using existing observational data alone, output both the discovered causal relationships and estimate the effect of actions simultaneously.\n \nCausica is a deep learning library for end-to-end causal inference, including both causal discovery and inference.  It implements deep end-to-end inference framework [2] and different alternatives.\n \nThis project splits the interventional decision making from observational decision making Azua repo found here [Azua](https://github.com/microsoft/project-azua).\n\nThis codebase has been heavily refactored, you can find the previous version of the code [here](https://github.com/microsoft/causica/releases/tag/v0.0.0).\n\n[1] Alvarez et al. [Simultaneous Missing Value Imputation and Structure Learning with Groups](https://openreview.net/pdf?id=4rm6tzBjChe), NeurIPS 2022\n\n[2] Geffner et al. [Deep End-to-end Causal Inference.](https://arxiv.org/pdf/2202.02195.pdf)\n\n[3] Gong et al.  [Rhino: Deep Causal Temporal Relationship Learning With History-dependent Noise](https://openreview.net/pdf?id=i_1rbq8yFWC), ICLR 2023\n\n[4] Ma et al. [Causal Reasoning in the Presence of Latent Confounders via Neural ADMG Learning]( https://openreview.net/pdf?id=dcN0CaXQhT), ICLR 2023\n\n# DECI: End to End Causal Inference\n\n## About\n\nReal-world data-driven decision making requires causal inference to ensure the validity of drawn conclusions. However, it is very uncommon to have a-priori perfect knowledge of the causal relationships underlying relevant variables. DECI allows the end user to perform causal inference without having complete knowledge of the causal graph. This is done by combining the causal discovery and causal inference steps in a single model. DECI takes in observational data and outputs ATE and CATE estimates. \n\nFor more information, please refer to the [paper](https://arxiv.org/abs/2202.02195).\n\n\n**Model Description**\n\nDECI is a generative model that employs an additive noise structural equation model (ANM-SEM) to capture the functional relationships among variables and exogenous noise, while simultaneously learning a variational distribution over causal graphs. Specifically, the relationships among variables are captured with flexible neural networks while the exogenous noise is modelled as either a Gaussian or spline-flow noise model. The SEM is reversible, meaning that we can generate an observation vector from an exogenous noise vector through forward simulation and given a observation vector we can recover a unique corresponding exogenous noise vector. In this sense, the DECI SEM can be seen as a flow from exogenous noise to observations. We employ a mean-field approximate posterior distribution over graphs, which is learnt together with the functional relationships among variables by optimising an evidence lower bound (ELBO). Additionally, DECI supports learning under partially observed data.\n\n**Simulation-based Causal Inference**\n\nDECI estimates causal quantities (ATE) by applying the relevant interventions to its learnt causal graph (i.e. mutilating incoming edges to intervened variables) and then sampling from the generative model. This process involves first sampling a vector of exogenous noise from the learnt noise distribution and then forward simulating the SEM until an observation vector is obtained. ATE can be computed via estimating an expectation over the effect variable of interest using MonteCarlo samples of the intervened distribution of observations. \n\n## How to run\n\nThe command to train this model the csuite_weak_arrows dataset and evaluate causal discovery and ATE estimation performance is:\n\n```\npython train_csuite_example.py --dataset csuite_weak_arrows\n```\n\nfrom within `examples/csuite_example`.\n\nThis will download the data from the CSuite Azure blob storage and train DECI on it. See [here](https://github.com/microsoft/csuite) for more info about CSuite datasets. The script will work on any of the available CSuite datasets.\n\nTo evaluate the model run:\n\n```\npython eval_csuite_example.py\n```\n\nThis should print various causal discovery and causal inference metrics.\n\n**Specifying a noise model**\n\nThe noise exogenous model can be modified by changing the `noise_dist` field within `training_default.json`, either \'gaussian\' or \'spline\' are allowed.\n\nThe Gaussian model has Gaussian exogenous noise distribution with mean set to 0 while its variance is learnt.\n\nThe Spline model uses a flexible spline flow that is learnt from the data. This model provides most gains in heavy-tailed noise settings, where the Gaussian model is at risk of overfitting to outliers, but can take longer to train.\n\n## Further extensions \n\nFor now, we have removed Rhino and DDECI from the codebase but they will be added back. You can still access the previously released versions [here](https://github.com/microsoft/causica/releases/tag/v0.0.0).\n\n# References\n\nIf you have used the models in our code base, please consider to cite the corresponding papers:\n\n[1], **(VISL)** Pablo Morales-Alvarez, Wenbo Gong, Angus Lamb, Simon Woodhead, Simon Peyton Jones, Nick Pawlowski, Miltiadis Allamanis, Cheng Zhang, "Simultaneous Missing Value Imputation and Structure Learning with Groups", [ArXiv preprint](https://arxiv.org/abs/2110.08223)\n\n[2], **(DECI)** Tomas Geffner, Javier Antoran, Adam Foster, Wenbo Gong, Chao Ma, Emre Kiciman, Amit Sharma, Angus Lamb, Martin Kukla, Nick Pawlowski, Miltiadis Allamanis, Cheng Zhang. Deep End-to-end Causal Inference. [Arxiv preprint](https://arxiv.org/abs/2202.02195) (2022)\n\n[3], **(DDECI)** Matthew Ashman, Chao Ma, Agrin Hilmkil, Joel Jennings, Cheng Zhang. Causal Reasoning in the Presence of Latent Confounders via Neural ADMG Learning. [ICLR](https://openreview.net/forum?id=dcN0CaXQhT) (2023)\n\n[4], **(Rhino)** Wenbo Gong, Joel Jennings, Cheng Zhang, Nick Pawlowski. Rhino: Deep Causal Temporal Relationship Learning with History-dependent Noise. [ICLR](https://openreview.net/forum?id=i_1rbq8yFWC) (2023)\n\nOther references:\n- Louizos, Christos, et al. "Causal effect inference with deep latent-variable models." Advances in neural information processing systems 30 (2017).\n- Hill, Jennifer L. "Bayesian nonparametric modeling for causal inference." Journal of Computational and Graphical Statistics 20.1 (2011): 217-240.\n\n\n# Development\n\n## Poetry\n\nWe use Poetry to manage the project dependencies, they\'re specified in the [pyproject.toml](pyproject.toml). To install poetry run:\n\n```\n    curl -sSL https://install.python-poetry.org | python3 -\n```\n\nTo install the environment run `poetry install`, this will create a virtualenv that you can use by running either `poetry shell` or `poetry run {command}`. It\'s also a virtualenv that you can interact with in the normal way too.\n\nMore information about poetry can be found [here](https://python-poetry.org/)\n\n## mlflow\n\nWe use [mlflow](https://mlflow.org/) for logging metrics and artifacts. By default it will run locally and store results in `./mlruns`.\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from enum import Enum
from typing import Dict, Optional, Tuple

import torch
from torch.nn import Parameter, ParameterDict

from causica.distributions import (
    NoiseAccessibleBernoulli,
    NoiseAccessibleCategorical,
    NoiseAccessibleIndependent,
    NoiseAccessibleMultivariateNormal,
    ParametrizedDistribution,
    SplineDistribution,
    create_spline_dist_params,
)
from causica.functional_relationships import ICGNN


class NoiseDist(Enum):
    SPLINE = "spline"
    GAUSSIAN = "gaussian"


class TrainableContainer(torch.nn.Module):
    """
    A module to contain all of the submodules associated with training on the CSuite example.

    This allows the state of the computation to be easily saved and loaded.
    """

    def __init__(
        self,
        dataset_name: str,
        icgnn: ICGNN,
        vardist: ParametrizedDistribution,
        noise_dist_params: ParameterDict,
        noise_dist_type: NoiseDist,
    ):
        """
        Args:
            dataset_name: the name of the CSuite dataset to train on
            icgnn: the ICGNN module
            vardist: The variational distribution that approximates the posterior
            noise_dist: The module of the parameters of the noise distribution for the SEM
            noise_dist_type: The continuous noise distribution used
        """
        super().__init__()
        self.dataset_name = dataset_name
        self.icgnn = icgnn
        self.vardist = vardist
        self.noise_dist = noise_dist_params
        self.noise_dist_type = noise_dist_type


def create_noise_dists(
    shapes: Dict[str, torch.Size],
    types_dict: Dict[str, str],
    noise_dist: NoiseDist,
    noise_dist_params: Optional[ParameterDict] = None,
) -> Tuple[Dict, ParameterDict]:
    """Create the noise distribution for each node along with its associated parameters."""
    param_dict = ParameterDict({}) if noise_dist_params is None else noise_dist_params
    noise_dist_funcs = {}
    for key, shape_tuple in shapes.items():
        shape = shape_tuple[-1]
        var_type = types_dict[key]
        if var_type == "categorical":
            if key not in param_dict:
                param_dict[key] = ParameterDict({"base_logits": Parameter(torch.zeros(shape), requires_grad=True)})
            noise_dist_funcs[key] = lambda x, params=param_dict[key]: NoiseAccessibleCategorical(
                delta_logits=x, **params
            )
        elif var_type == "binary":
            if key not in param_dict:
                param_dict[key] = ParameterDict({"base_logits": Parameter(torch.zeros(shape), requires_grad=True)})
            noise_dist_funcs[key] = lambda x, params=param_dict[key]: NoiseAccessibleIndependent(
                NoiseAccessibleBernoulli(delta_logits=x, **params), 1
            )
        # continuous variables
        elif noise_dist == NoiseDist.SPLINE:
            if key not in param_dict:
                noise_param_list = create_spline_dist_params(features=shape, flow_steps=1)
                base_scale = Parameter(torch.ones(shape), requires_grad=False)
                base_loc = Parameter(torch.zeros(shape), requires_grad=False)
                param_dict[key] = ParameterDict(
                    dict(base_loc=base_loc, base_scale=base_scale, param_list=noise_param_list)
                )
            noise_dist_funcs[key] = lambda x, params=param_dict[key]: NoiseAccessibleIndependent(
                SplineDistribution(output_bias=x, **params), 1
            )
        elif noise_dist == NoiseDist.GAUSSIAN:
            if key not in param_dict:
                param_dict[key] = ParameterDict({"log_scale": torch.zeros([shape], requires_grad=True)})
            noise_dist_funcs[key] = lambda x, params=param_dict[key]: NoiseAccessibleMultivariateNormal(
                loc=x, scale_tril=torch.diag(torch.exp(params["log_scale"]))
            )
        else:
            raise ValueError("Unrecognised model", noise_dist)

    return noise_dist_funcs, param_dict



from causica.lightning.cli import LightningCLIWithDefaults
from causica.lightning.data_modules import CSuiteDataModule
from causica.lightning.modules import DECIModule

if __name__ == "__main__":
    LightningCLIWithDefaults(DECIModule, CSuiteDataModule, run=True, save_config_callback=None)



import argparse
import dataclasses
import math
from typing import Dict, Optional

import mlflow
import pytorch_lightning as pl
import torch
from tensordict import TensorDict

from causica.datasets.csuite_data import (
    CSUITE_DATASETS_PATH,
    CounterfactualsWithEffects,
    DataEnum,
    InterventionsWithEffects,
    get_categorical_sizes,
    load_data,
)
from causica.datasets.interventional_data import CounterfactualData, InterventionData
from causica.datasets.tensordict_utils import convert_one_hot, tensordict_shapes
from causica.functional_relationships import ICGNN
from causica.graph.evaluation_metrics import adjacency_f1, orientation_f1
from causica.sem.distribution_parameters_sem import DistributionParametersSEM
from causica.training.evaluation import (
    eval_ate_rmse,
    eval_intervention_likelihoods,
    eval_ite_rmse,
    list_logsumexp,
    list_mean,
)
from causica.training.trainable_container import TrainableContainer, create_noise_dists

NUM_GRAPH_SAMPLES = 100
NUM_ATE_ITE_SEMS = 10
LAST_MODEL = "last_model.pt"


@dataclasses.dataclass
class EvaluationMetrics:
    adjacency_f1: float
    orientation_f1: float
    test_log_likelihood: float
    interventional_log_likelihood: float
    ate_rmse: float
    ite_rmse: Optional[float] = None


def eval_csuite(container: TrainableContainer, ite: bool = False) -> EvaluationMetrics:
    device = "cpu"
    container.to(device)

    true_adj_matrix = torch.tensor(
        load_data(CSUITE_DATASETS_PATH, container.dataset_name, DataEnum.TRUE_ADJACENCY), device=device
    )

    icgnn: ICGNN = container.icgnn
    node_names = list(icgnn.variables.keys())

    vardist = container.vardist.forward()
    graph_samples = vardist.sample(torch.Size([NUM_GRAPH_SAMPLES]))

    adj_f1 = list_mean([adjacency_f1(true_adj_matrix, graph) for graph in graph_samples]).item()
    orient_f1 = list_mean([orientation_f1(true_adj_matrix, graph) for graph in graph_samples]).item()

    variables_metadata = load_data(CSUITE_DATASETS_PATH, container.dataset_name, DataEnum.VARIABLES_JSON)

    categorical_sizes = get_categorical_sizes(variables_list=variables_metadata["variables"])

    def one_hotify(data: TensorDict) -> TensorDict:
        return convert_one_hot(data, categorical_sizes)

    dataset_test: TensorDict = one_hotify(
        load_data(CSUITE_DATASETS_PATH, container.dataset_name, DataEnum.TEST, variables_metadata)
    )
    dataset_test = dataset_test.apply(lambda t: t.to(dtype=torch.float32, device=device))
    variables = variables_metadata["variables"]
    types_dict = {var["group_name"]: var["type"] for var in variables}
    noise_dist_funcs, _ = create_noise_dists(
        shapes=tensordict_shapes(dataset_test),
        types_dict=types_dict,
        noise_dist=container.noise_dist_type,
        noise_dist_params=container.noise_dist,
    )

    # use an iterator to minimize the memory impact
    sems = (
        DistributionParametersSEM(graph=graph, node_names=node_names, noise_dist=noise_dist_funcs, func=icgnn)
        for graph in graph_samples
    )

    # estimate total log prob using graph samples and report the mean over graphs
    log_prob = torch.mean(
        list_logsumexp([sem.log_prob(dataset_test) for sem in sems]) - math.log(NUM_GRAPH_SAMPLES)
    ).item()

    interventions: InterventionsWithEffects = load_data(
        CSUITE_DATASETS_PATH, container.dataset_name, DataEnum.INTERVENTIONS, variables_metadata
    )

    def intersect(dict_1: Dict, dict_2: Dict):
        return {key: dict_1[key] for key in dict_1.keys() & dict_2.keys()}

    def one_hot_intervention(intervention: InterventionData) -> InterventionData:
        return InterventionData(
            convert_one_hot(
                intervention.intervention_data, intersect(categorical_sizes, intervention.intervention_data)
            ),
            convert_one_hot(
                intervention.intervention_values, intersect(categorical_sizes, intervention.intervention_values)
            ),
            convert_one_hot(intervention.condition_values, intersect(categorical_sizes, intervention.condition_values)),
        )

    interventions = [
        (one_hot_intervention(int_a), one_hot_intervention(int_b), effects) for int_a, int_b, effects in interventions
    ]

    graph_samples = vardist.sample(torch.Size([NUM_ATE_ITE_SEMS]))
    sems_list = [
        DistributionParametersSEM(graph=graph, node_names=node_names, noise_dist=noise_dist_funcs, func=icgnn)
        for graph in graph_samples
    ]

    interventional_log_prob = eval_intervention_likelihoods(sems_list, interventions).item()

    mean_ate_rmse = eval_ate_rmse(sems_list, interventions).item()

    if ite:
        counterfactuals: CounterfactualsWithEffects = load_data(
            CSUITE_DATASETS_PATH, container.dataset_name, DataEnum.COUNTERFACTUALS, variables_metadata
        )

        def one_hot_counterfactual(counterfactual: CounterfactualData) -> CounterfactualData:
            return CounterfactualData(
                convert_one_hot(
                    counterfactual.counterfactual_data, intersect(categorical_sizes, counterfactual.counterfactual_data)
                ),
                convert_one_hot(
                    counterfactual.intervention_values, intersect(categorical_sizes, counterfactual.intervention_values)
                ),
                convert_one_hot(counterfactual.factual_data, intersect(categorical_sizes, counterfactual.factual_data)),
            )

        counterfactuals = [
            (one_hot_counterfactual(c_a), one_hot_counterfactual(c_b), effects) for c_a, c_b, effects in counterfactuals
        ]
        mean_ite_rmse = eval_ite_rmse(sems_list, counterfactuals).item()
    else:
        mean_ite_rmse = None

    return EvaluationMetrics(
        adjacency_f1=adj_f1,
        orientation_f1=orient_f1,
        test_log_likelihood=log_prob,
        interventional_log_likelihood=interventional_log_prob,
        ate_rmse=mean_ate_rmse,
        ite_rmse=mean_ite_rmse,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model-save-file", "-s", default=LAST_MODEL)
    parser.add_argument("--ite", default=False, action="store_true")

    args = parser.parse_args()

    pl.seed_everything(123)

    train_container: TrainableContainer = torch.load(args.model_save_file)["container"]
    eval_metrics = eval_csuite(train_container, args.ite)

    print(eval_metrics)
    mlflow.log_metric("eval/adjacency.f1", eval_metrics.adjacency_f1)
    mlflow.log_metric("eval/orientation.f1", eval_metrics.orientation_f1)
    mlflow.log_metric("eval/test_LL", eval_metrics.test_log_likelihood)
    mlflow.log_metric("eval/Interventional_LL", eval_metrics.interventional_log_likelihood)
    mlflow.log_metric("eval/ATE_RMSE", eval_metrics.ate_rmse)
    if eval_metrics.ite_rmse is not None:
        mlflow.log_metric("eval/ITE_RMSE", eval_metrics.ite_rmse)



from .bernoulli import NoiseAccessibleBernoulli
from .categorical import NoiseAccessibleCategorical
from .multivariate_normal import NoiseAccessibleMultivariateNormal
from .noise_accessible import NoiseAccessible, NoiseAccessibleDistribution, NoiseAccessibleIndependent



import abc

import torch
import torch.distributions as td


class NoiseAccessible(abc.ABC):
    """
    Extend Distributions to allow the noise (usually unparametrized) to be extracted from samples and vice versa.

    This class is used for Counterfactuals, where we want to keep the noise fixed but fix some values of the SEM.
    """

    @abc.abstractmethod
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        pass

    @abc.abstractmethod
    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        pass


class NoiseAccessibleDistribution(NoiseAccessible, td.Distribution):
    """
    A class used for type hinting.

    Do not inherit from this class. Instead, inherit from the distribution you're interested in
    and mixin the noise accessible class.
    """


class NoiseAccessibleIndependent(td.Independent, NoiseAccessible):
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        This just passes through to the underlying distribution.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return self.base_dist.sample_to_noise(samples)

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        This just passes through to the underlying distribution.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """

        return self.base_dist.noise_to_sample(noise)
